from trame.internal.app import get_app_instance
from trame.html import AbstractElement
from pywebvue.modules import XAI

# Make sure used module is available
_app = get_app_instance()
_app.enable_module(XAI)


class XaiHeatMap(AbstractElement):
    def __init__(self, children=None, **kwargs):
        super().__init__("xai-heat-map", children, **kwargs)
        self._attr_names += [
            "heatmap",
            "shape",
            "color_mode",
            "color_range",
            "color_preset",
        ]
        self._event_names += [
            "hover",
            "enter",
        ]


class XaiImage(AbstractElement):
    def __init__(self, children=None, **kwargs):
        super().__init__("xai-image", children, **kwargs)
        self._attr_names += [
            "src",
            "max_height",
            "max_width",
            "width",
            "colors",
            "areas",
            "area_key",
            "area_style",
            "area_selected",
            "area_selected_opacity",
            "area_opacity",
            "heatmaps",
            "heatmap_opacity",
            "heatmap_color_preset",
            "heatmap_color_range",
            "heatmap_active",
            "heatmap_color_mode",
        ]
        self._event_names += [
            ("area_selection_change", "areaSelectionChange"),
            ("color_range", "colorRange"),
            "hover",
            "enter",
        ]



def get_cli_parser():
    """Run or add args to CLI parser

    :returns: Parser from argparse

    >>> parser = get_cli_parser()
    >>> parser.add_argument("-o", "--output", help="Working directory")
    >>> args, unknown = parser.parse_known_args()
    >>> print(args.output)
    """
    from trame.internal.app import get_app_instance

    _app = get_app_instance()
    return _app.cli_parser



try:
    # Use importlib metadata if available
    from importlib.metadata import version, PackageNotFoundError
except ImportError:
    # No importlib metadata. Try to use pkg_resources instead.
    from pkg_resources import (
        get_distribution,
        DistributionNotFound as PackageNotFoundError,
    )

    def version(x):
        return get_distribution(x).version


def get_version():
    try:
        return version("trame")
    except PackageNotFoundError:
        # package is not installed
        pass



import asyncio
from multiprocessing import Process
from threading import Thread, current_thread


class AppServerThread(Thread):
    def __init__(self, app, port=None, **kwargs):
        Thread.__init__(self, **kwargs)
        self._app = app
        self._port = port
        self._loop = None
        self._thread_id = None

    def run(self):
        self._loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self._loop)
        self._thread_id = current_thread()
        self._app.run_server(port=self._port, thread=True)

    def stop(self):
        self._loop.call_soon_threadsafe(self._app.stop_server)

    @property
    def port(self):
        return self._app.server_port


class ClientWindowProcess(Process):
    def __init__(
        self, title=None, port=None, msg_queue=None, file_dialog=None, **kwargs
    ):
        Process.__init__(self)
        self._title = title
        self._port = port
        self._msg_queue = msg_queue
        self._window_args = kwargs
        self._file_dialog = file_dialog
        self._main_window = None

    def _open_file_dialog(self):
        result = self._main_window.create_file_dialog(**self._file_dialog)
        self._msg_queue.put(["file_dialog", result])

    def exit(self):
        self._main_window.destroy()
        self._msg_queue.put("closing")

    def run(self):
        try:
            import webview
        except ImportError:
            print("trame.start_desktop_window() require pywebview==3.4")
            return

        self._main_window = webview.create_window(
            title=self._title,
            url=f"http://localhost:{self._port}/",
            **self._window_args,
        )
        self._main_window.closing += self.exit

        if self._file_dialog:
            webview.start(func=self._open_file_dialog)
        else:
            webview.start()



import types

import trame.internal as tri


def str_key_prefix(txt):
    if txt.startswith("{") or txt.startswith("`"):
        return ":"

    return ""


def py2js_key(key):
    return key.replace("_", "-")


class ElementContextManager:
    def __init__(self):
        self.element_stack = []

    def enter(self, elem):
        self.element_stack.append(elem)

    def exit(self, elem):
        if len(self.element_stack) and elem == self.element_stack[-1]:
            self.element_stack.pop()

    def add_child(self, elem):
        if len(self.element_stack):
            self.element_stack[-1].add_child(elem)


HTML_CTX = ElementContextManager()


class AbstractElement:
    """
    A Vue component which can integrate with the rest of trame

    See Vue docs |vue_doc_link| for more info

    .. |vue_doc_link| raw:: html

        <a href="https://vuejs.org/v2/guide/instance.html" target="_blank">here</a>

    .. |mdn_doc_link| raw:: html

        <a href="https://developer.mozilla.org/en-US/docs/Web/HTML/Attributes" target="_blank">here</a>

    .. |mdn_event_link| raw:: html

        <a href="https://developer.mozilla.org/en-US/docs/Web/API/Element#mouse_events">here</a>

    :param name: The name of the element, like 'div' for a ``<div/>`` element
    :type name: str
    :param children: The children nested within this element
    :type children:  str | list[trame.html.*] | trame.html.* | None
    :param __properties: Provide more attribute names that should be handle
    :param __events: Provide more event names that should be handle

    Html attributes - See |mdn_doc_link| for more info

    :param id: See |mdn_doc_link| for more info
    :param classes: Match the HTML `class` attribute. See |mdn_doc_link| for more info
    :param style: See |mdn_doc_link| for more info

    Vue attributes - See |vue_doc_link| for more info

    :param ref: See |vue_doc_link| for more info
    :param v_model: See |vue_doc_link| for more info
    :param v_if: See |vue_doc_link| for more info
    :param v_show: See |vue_doc_link| for more info
    :param v_for: See |vue_doc_link| for more info
    :param key: See |vue_doc_link| for more info

    Events - See |mdn_event_link| for more info

    :param click: See |mdn_event_link| for more info
    :param mousedown: See |mdn_event_link| for more info
    :param mouseup: See |mdn_event_link| for more info
    :param mouseenter: See |mdn_event_link| for more info
    :param mouseleave: See |mdn_event_link| for more info
    :param contextmenu: See |mdn_event_link| for more info

    """

    _next_id = 1

    def __init__(self, _elem_name, children=None, **kwargs):
        AbstractElement._next_id += 1
        self._id = AbstractElement._next_id
        self._elem_name = _elem_name
        self._allowed_keys = set()
        self._attr_names = kwargs.get("__properties", [])
        self._event_names = kwargs.get("__events", [])

        self._attributes = {}
        self._py_attr = kwargs
        self._children = []

        if children:
            if isinstance(children, list):
                self._children.extend(children)
            else:
                self._children.append(children)

        # Add standard Vue attr/event handling
        self._attr_names += [
            "id",
            "ref",
            ("classes", "class"),
            "style",
            ("key", ":key"),
            # default vue.js directives
            "v_text",
            "v_html",
            "v_show",
            "v_if",
            "v_else",
            "v_else_if",
            "v_for",
            "v_model",
            "v_pre",
            "v_once",
        ]
        self._event_names += [
            "click",
            "mousedown",
            "mouseup",
            "mouseenter",
            "mouseleave",
            "contextmenu",
        ]

        # Add ourself to context if any
        HTML_CTX.add_child(self)

    def _attr_str(self):
        return " ".join(self._attributes.values())

    def _update_allowed_keys(self):
        if hasattr(self, "_attr_names") and hasattr(self, "_event_names"):
            for items in [self._attr_names, self._event_names]:
                for item in items:
                    if isinstance(item, str):
                        self._allowed_keys.add(item)
                    else:
                        self._allowed_keys.add(item[0])

    # -------------------------------------------------------------------------
    # Buildin API
    # -------------------------------------------------------------------------

    def __getitem__(self, name):
        return self._py_attr[name]

    def __setitem__(self, name, value):
        if name in self._allowed_keys:
            self._py_attr[name] = value
        else:
            print(f"Attribute {name} is not defined for {self._elem_name}")

    def __getattr__(self, name):
        if name[0] == "_":
            raise AttributeError()

        return self._py_attr[name]

    def __setattr__(self, name, value):
        if name[0] == "_":
            self.__dict__[name] = value
        elif name == "children":
            self._children = value
        elif name in self._allowed_keys:
            self._py_attr[name] = value
        else:
            self.__dict__[name] = value

        if name in ["_attr_names", "_event_names"]:
            self._update_allowed_keys()

    # -------------------------------------------------------------------------
    # helpers
    # -------------------------------------------------------------------------

    def ttsSensitive(self):
        """
        Calling this function on an element will make it fully recreate itself
        every time the layout update. Internally it is managed by adding a `key=`
        attribute which use a layout timestamp.

        This is especially useful for component that manage other elements outside
        of themself like VSelect in Vuetify.
        """
        self._attributes["__tts"] = f':key="`w{self._id}-${{tts}}`"'
        return self

    def attrs(self, *names):
        """
        Calling this function will process the provided attribute names and
        configure its internal so the macthing HTML string could easily be
        generated later on.

        :param names: The names attribute to process
        :type names: *str
        """
        _app = tri.get_app_instance()
        for _name in names:
            js_key = None
            name = _name
            if isinstance(_name, tuple):
                name = _name[0]
                js_key = _name[1]

            if name in self._py_attr:
                if js_key is None:
                    js_key = py2js_key(name)
                value = self._py_attr[name]

                if value is None:
                    continue

                if (
                    _app._debug
                    and js_key.startswith("v-")
                    and not isinstance(value, (tuple, list))
                ):
                    print(
                        f'Warning: A Vue directive is evaluating your expression and trame would expect a tuple instead of a plain type. <{self._elem_name} {js_key}="{value}" ... />'
                    )

                if isinstance(value, (tuple, list)):
                    if len(value) > 1 and value[0] not in _app.state:
                        _app.state[value[0]] = value[1]

                    if js_key.startswith("v-"):
                        self._attributes[name] = f'{js_key}="{value[0]}"'
                    elif js_key.startswith(":"):
                        self._attributes[name] = f'{js_key}="{value[0]}"'
                    else:
                        self._attributes[name] = f':{js_key}="{value[0]}"'
                elif isinstance(value, bool):
                    if value:
                        self._attributes[name] = js_key
                elif isinstance(value, (str, int, float)):
                    self._attributes[name] = f'{js_key}="{value}"'
                else:
                    print(
                        f"Error: Don't know how to handle attribue name '{name}' with value '{value}' in {self.__class__}::{self._elem_name}"
                    )

        return self

    def events(self, *names):
        """
        Calling this function will process the provided event names and
        configure its internal so the macthing HTML string could easily be
        generated later on.

        :param names: The names events to process
        :type names: *str
        """
        _app = tri.get_app_instance()
        for _name in names:
            js_key = None
            name = _name
            if isinstance(_name, tuple):
                name = _name[0]
                js_key = _name[1]
            if name in self._py_attr:
                if js_key is None:
                    js_key = py2js_key(name)
                js_key = f"@{js_key}"
                value = self._py_attr[name]

                if value is None:
                    continue

                if isinstance(value, str):
                    self._attributes[name] = f'{js_key}="{value}"'
                elif isinstance(value, (types.FunctionType, types.MethodType)):
                    trigger_name = tri.trigger_key(value)
                    self._attributes[name] = f"{js_key}=\"trigger('{trigger_name}')\""
                elif isinstance(value, tuple):
                    trigger_name = value[0]
                    if isinstance(trigger_name, (types.FunctionType, types.MethodType)):
                        trigger_name = tri.trigger_key(trigger_name)
                    if len(value) == 1:
                        self._attributes[
                            name
                        ] = f"{js_key}=\"trigger('{trigger_name}')\""
                    if len(value) == 2:
                        self._attributes[
                            name
                        ] = f"{js_key}=\"trigger('{trigger_name}', {value[1]})\""
                    if len(value) == 3:
                        self._attributes[
                            name
                        ] = f"{js_key}=\"trigger('{trigger_name}', {value[1]}, {value[2]})\""
                else:
                    print(
                        f"Error: Don't know how to handle event name '{name}' with value '{value}' in {self.__class__}::{self._elem_name}"
                    )
        return self

    def clear(self):
        """
        Remove all children
        """
        self._children.clear()

    def hide(self):
        """
        Hide element while keeping it in the DOM. (display: none)
        """
        self._attributes["__style"] = 'style="display: none"'

    def add_child(self, child):
        """
        Add a component to this component's children

        :param child: The component to add as a child
        :type child: str | AbstractElement
        """
        self._children.append(child)

    def add_children(self, children):
        """
        Add components to this component's children.
        The provided children is expected to be a list.

        :param children: The list of components to add to the children
        :type children: list
        """
        self._children += children

    @property
    def children(self):
        """
        Children components
        """
        return self._children

    def set_text(self, value):
        """
        Replace children with a single text child element

        :param value: The text for the new text child element
        :type value: str
        """
        self.clear()
        self._children.append(value)

    @property
    def html(self):
        """
        Return a string representation of the HTML component
        """
        # Build attributes
        self.attrs(*self._attr_names)
        self.events(*self._event_names)

        # Compute HTML str
        if len(self._children):
            out_buffer = []
            out_buffer.append(f"<{self._elem_name} {self._attr_str()}>")
            for child in self._children:
                if isinstance(child, str):
                    out_buffer.append(child)
                else:
                    out_buffer.append(child.html)
            out_buffer.append(f"</{self._elem_name}>")
            return "\n".join(out_buffer)
        else:
            return f"<{self._elem_name} {self._attr_str()} />"

    # -------------------------------------------------------------------------
    # Resource manager
    # -------------------------------------------------------------------------

    def __enter__(self):
        HTML_CTX.enter(self)
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        HTML_CTX.exit(self)


class Element(AbstractElement):
    """
    Any html element you would like to use in trame

    :param _elem_name: The name of the element, like 'div' for a ``<div/>`` element
    :type _elem_name: str
    :param children: The children nested within this element
    :type children:  str | list[trame.html.*] | trame.html.* | None
    """

    def __init__(self, _elem_name, children=None, **kwargs):
        super().__init__(_elem_name, children, **kwargs)


class Div(AbstractElement):
    """
    The standard html content div element

    :param children: The children nested within this element
    :type children:  str | list[trame.html.*] | trame.html.* | None
    """

    def __init__(self, children=None, **kwargs):
        super().__init__("div", children, **kwargs)


class Span(AbstractElement):
    """
    The standard html content span element

    :param children: The children nested within this element
    :type children:  str | list[trame.html.*] | trame.html.* | None
    """

    def __init__(self, children=None, **kwargs):
        super().__init__("span", children, **kwargs)


class Form(AbstractElement):
    """
    The standard html form element

    :param children: The children nested within this element
    :type children:  str | list[trame.html.*] | trame.html.* | None
    """

    def __init__(self, children=None, **kwargs):
        super().__init__("form", children, **kwargs)
        self._attr_names += ["action"]


class Label(AbstractElement):
    """
    The standard html input label element

    :param children: The children nested within this element
    :type children:  str | list[trame.html.*] | trame.html.* | None
    """

    def __init__(self, children=None, **kwargs):
        super().__init__("label", children, **kwargs)


class Input(AbstractElement):
    """
    The standard html input (form input) element

    :param children: The children nested within this element
    :type children:  str | list[trame.html.*] | trame.html.* | None
    """

    def __init__(self, children=None, **kwargs):
        super().__init__("input", children, **kwargs)
        self._attr_names += [
            "type",
            "value",
            "name",
            "size",
            "min",
            "max",
            "step",
            "maxlength",
            "disabled",
            "readonly",
            "multiple",
            "pattern",
            "placeholder",
            "required",
            "autofocus",
            "src",
            "width",
            "height",
            "alt",
            "list",
            "autocomplete",
        ]
        self._event_names += ["change", "input"]


class Template(AbstractElement):
    """
    The standard html content template element. This is mostly used by |slot_doc_link|.

    .. |slot_doc_link| raw:: html

        <a href="https://vuejs.org/v2/guide/instance.html" target="_blank">vue's slot system</a>


    :param children: The children nested within this element
    :type children:  str | list[trame.html.*] | trame.html.* | None
    :param v_slot: The slot this template corresponds to
    """

    slot_names = set()

    def __init__(self, children=None, **kwargs):
        super().__init__("template", children, **kwargs)
        self._attr_names += ["v_slot"]
        for slot_name in Template.slot_names:
            safe_name = slot_name.replace("-", "_").replace(".", "_")
            if "<name>" in safe_name:
                safe_header, safe_tail = safe_name.split("<name>")
                header, tail = slot_name.split("<name>")
                for key in kwargs:
                    if key.startswith(header):
                        dyna_name = key[len(header) : -len(tail)]
                        self._attr_names.append(
                            (
                                f"v_slot_{safe_header}{dyna_name}{safe_tail}",
                                f"v-slot:{header}{dyna_name}{tail}",
                            )
                        )
            else:
                self._attr_names.append((f"v_slot_{safe_name}", f"v-slot:{slot_name}"))


class StateChange(AbstractElement):
    """
    Component to react when a state entry change so an event can be triggered

    :param name: Which part of the state to listen to
    :type name: str

    Events

    :param change: Function to run if state changes
    :type change: function
    """

    def __init__(self, name, **kwargs):
        super().__init__("py-state-update", **kwargs)
        self._attributes["value"] = f':value="{name}"'
        self._event_names += [
            "change",
        ]


class Triggers(AbstractElement):
    """
    Component to trigger JS actions from Python

    :param ref: Name for Vue reference to this object
    :type ref: str
    :param triggers: Mapping from names of triggers to expressions or methods in JS which they will call
    :type triggers: dict[str, str]

    >>> triggers = trame.html.Triggers(ref="all_triggers", triggers={ "reset_camera": "$refs.view.resetCamera()" })

    """

    def __init__(self, ref, triggers={}, **kwargs):
        super().__init__("py-trigger", **kwargs)
        self._ref = ref
        self._attributes["ref"] = f'ref="{ref}"'
        for key, value in triggers.items():
            self._attributes[f"_{key}"] = f'@{key}="{value}"'

    def add(self, name, call):
        """
        Add a trigger which can call JS from Python

        :param name: Reference for this JS method or expression trigger
        :type name: str
        :param call: JS method or expression to call when triggered
        :type call: str

        >>> triggers.add("created", "console.log('UI is created')")
        >>> triggers.add("mounted", "console.log('UI is mounted')")
        >>> triggers.add("beforeDestroy", "console.log('UI is going away')")

        """
        self._attributes[f"_{name}"] = f'@{name}="{call}"'

    def call(self, name, *args):
        """
        Trigger JS code previously added to this object

        :param name: Reference for this JS method or expression trigger
        :type name: str
        :param args: Parameters passed to JS method

        >>> triggers.call("reset_camera")

        """
        _app = tri.get_app_instance()
        _app.update(ref=self._ref, method="emit", args=[name, *args])


class VTKLoading(AbstractElement):
    """
    Component to show the 3 spinning partial circles using the ParaView Red/Green/Yellow colors.

    :param message: Message to put below the spinning circles
    :type message: str
    """

    def __init__(self, message="", **kwargs):
        super().__init__("vtk-loading", message=message, **kwargs)
        self._attr_names += ["message"]



from trame.internal.utils import is_dunder
from .decorators import trigger


class Controller:
    """Controller acts as a container for function proxies

    It allows functions to be passed around that are not yet defined,
    and can be defined or re-defined later. For example:

    >>> f = ctrl.hello_func  # function is currently undefined
    >>> ctrl.hello_func = lambda: print("Hello, world!")
    >>> f()
    Hello, world!

    >>> ctrl.hello_func = lambda: print("Hello again!")
    >>> f()
    Hello again!
    """
    def __init__(self):
        super().__setattr__('_func_dict', {})

    def __getattr__(self, name):
        if is_dunder(name):
            return super().__getattr__(name)

        if name not in self._func_dict:
            self._func_dict[name] = ControllerFunction(name)

        # The ControllerFunction object is callable, but we will return the
        # __call__ method for greater compatibility in case a function is
        # required.
        return self._func_dict[name].__call__

    def __setattr__(self, name, func):
        # Do not allow pre-existing attributes, such as `trigger`, to be
        # re-defined.
        if name in self.__dict__ or name in Controller.__dict__:
            msg = (
                f"'{name}' is a special attribute on Controller that cannot "
                "be re-assigned"
            )
            raise Exception(msg)

        if name in self._func_dict:
            self._func_dict[name].func = func
        else:
            self._func_dict[name] = ControllerFunction(name, func)

    # Allow the trigger decorator to be accessed via @controller.trigger
    trigger = staticmethod(trigger)


class ControllerFunction:
    """Controller functions are callable function proxy objects

    Any calls are forwarded to the internal function, which may be
    undefined or dynamically changed. If a call is made when the
    internal function is undefined, a FunctionNotImplementedError is
    raised.
    """
    def __init__(self, name, func=None):
        # The name is needed to provide more helpful information upon
        # a FunctionNotImplementedError exception.
        self.name = name
        self.func = func

    def __call__(self, *args, **kwargs):
        if self.func is None:
            raise FunctionNotImplementedError(self.name)

        return self.func(*args, **kwargs)


class FunctionNotImplementedError(Exception):
    pass



from trame.internal.app import get_app_instance
from trame.html import AbstractElement

MODULE = None


def use_module(m):
    global MODULE
    MODULE = m
    _app = get_app_instance()
    _app.enable_module(m)


class VtkAlgorithm(AbstractElement):
    def __init__(self, children=None, **kwargs):
        super().__init__("vtk-algorithm", children, **kwargs)
        self._attr_names += ["port", "vtk_class", "state"]


class VtkCellData(AbstractElement):
    def __init__(self, children=None, **kwargs):
        super().__init__("vtk-cell-data", children, **kwargs)


class VtkDataArray(AbstractElement):
    def __init__(self, **kwargs):
        super().__init__("vtk-data-array", **kwargs)
        self._attr_names += [
            "name",
            "registration",
            "type",
            "values",
            "number_of_components",
        ]


class VtkFieldData(AbstractElement):
    def __init__(self, children=None, **kwargs):
        super().__init__("vtk-field-data", children, **kwargs)


class VtkGeometryRepresentation(AbstractElement):
    def __init__(self, children=None, **kwargs):
        super().__init__("vtk-geometry-representation", children, **kwargs)
        self._attr_names += [
            "id",
            "color_map_preset",
            "color_data_range",
            "actor",
            "mapper",
            "property",
        ]


class VtkGlyphRepresentation(AbstractElement):
    def __init__(self, children=None, **kwargs):
        super().__init__("vtk-glyph-representation", children, **kwargs)
        self._attr_names += [
            "color_map_preset",
            "color_data_range",
            "actor",
            "mapper",
            "property",
        ]


class VtkMesh(AbstractElement):
    def __init__(
        self,
        name,
        dataset=None,
        field_to_keep=None,
        point_arrays=None,
        cell_arrays=None,
        **kwargs,
    ):
        super().__init__("vtk-mesh", **kwargs)
        self.__name = name
        self.__dataset = dataset
        self.__field_to_keep = field_to_keep
        self.__point_arrays = point_arrays
        self.__cell_arrays = cell_arrays
        self._attr_names += ["port", "state"]
        if dataset:
            self._attributes["state"] = f':state="{name}"'
            self.update()

    def set_dataset(self, dataset):
        """
        Change this mesh's internal dataset and update shared state"""
        self.__dataset = dataset
        self.update()

    def update(self, **kwargs):
        """
        Propagate changes in internal data to shared state
        """
        if self.__dataset:
            _app = get_app_instance()
            _app.set(
                self.__name,
                MODULE.mesh(
                    self.__dataset,
                    field_to_keep=kwargs.get("field_to_keep", self.__field_to_keep),
                    point_arrays=kwargs.get("point_arrays", self.__point_arrays),
                    cell_arrays=kwargs.get("cell_arrays", self.__cell_arrays),
                ),
            )


class VtkPointData(AbstractElement):
    def __init__(self, children=None, **kwargs):
        super().__init__("vtk-point-data", children, **kwargs)


class VtkPolyData(AbstractElement):
    def __init__(self, name, children=None, dataset=None, **kwargs):
        super().__init__("vtk-polydata", children, **kwargs)
        self.__name = name
        self.__dataset = dataset
        self._attr_names += [
            "port",
            "verts",
            "lines",
            "polys",
            "strips",
            "connectivity",
        ]
        if dataset:
            self._attributes["bind"] = f'v-bind="{name}.mesh"'
            self.update()

    def set_dataset(self, dataset):
        """
        Change this polydata's internal dataset and update shared state
        """
        self.__dataset = dataset
        self.update()

    def update(self):
        """
        Propagate changes in internal data to shared state
        """
        if self.__dataset:
            _app = get_app_instance()
            _app.set(self.__name, MODULE.mesh(self.__dataset))


class VtkReader(AbstractElement):
    def __init__(self, **kwargs):
        super().__init__("vtk-reader", **kwargs)
        self._attr_names += [
            "parse_as_array_buffer",
            "parse_as_text",
            "port",
            "render_on_update",
            "reset_camera_on_update",
            "url",
            "vtk_class",
        ]


class VtkRemoteLocalView(AbstractElement):
    """
    The VtkRemoteLocalView component is a blend of VtkLocalView and VtkRemoteView where the user can choose dynamically which mode they want to be in. When instantiating a VtkRemoteLocalView several variables and triggers will be created for you to more easily control your view.

    >>> rl_view = vtk.VtkRemoteLocalView(
    ...   view=...,                # Instance of the view (required)
    ...                            # - VTK: vtkRenderWindow
    ...                            # - Paraview: viewProxy
    ...   # Just VtkRemoteLocalView params
    ...   namespace=...,           # Prefix for variables and triggers. See below. (required)
    ...   mode="local",            # Decide between local or remote. See below.
    ...
    ...   # VtkRemoteView params
    ...   **remote_view_params,
    ...
    ...   # VtkLocalView params
    ...   **local_view_params,
    ... )
    """

    def __init__(self, view, enable_rendering=True, **kwargs):
        super().__init__("vtk-remote-local-view", **kwargs)
        self.__app = get_app_instance()
        __ns = kwargs.get("namespace", "view")
        self.__mode_key = f"{__ns}Mode"
        self.__scene_id = f"{__ns}Scene"
        self.__ref = kwargs.get("ref", __ns)
        self.__rendering = enable_rendering

        # !!! HACK !!!
        # Allow user to configure view mode by providing (..., local/remote) and or "local/remote"
        __mode_expression = kwargs.get("mode", self.__mode_key)
        __mode_start = "remote"
        if isinstance(__mode_expression, (tuple, list)):
            if len(__mode_expression) == 2:
                __mode_expression, __mode_start = __mode_expression
            else:
                __mode_expression = __mode_expression[0]
        elif __mode_expression in ["local", "remote"]:
            __mode_start = __mode_expression
            __mode_expression = self.__mode_key

        self._attributes["mode"] = f':mode="{__mode_expression}"'
        # !!! HACK !!!

        self.__view_id = MODULE.id(view)
        self.__view = view
        self.__wrapped_view = MODULE.view(view, name=__ns, mode=__mode_start)

        # Provide mandatory attributes
        self._attributes["wsClient"] = ':wsClient="wsClient"'
        self._attributes["ref"] = f'ref="{self.__ref}"'
        self._attributes["view_id"] = f'id="{self.__view_id}"'
        self._attributes["view_state"] = f':viewState="{self.__scene_id}"'
        self._attributes["namespace"] = f'namespace="{__ns}"'

        self._attr_names += [
            # "mode", # <--- Managed by hand above
            "context_name",
            "interactive_ratio",
            "interactor_events",
            "interactor_settings",
        ]
        self._event_names += kwargs.get("interactor_events", [])

    def update_geometry(self):
        """
        Force update to geometry
        """
        self.__app.set(self.__scene_id, MODULE.scene(self.__view))

    def update_image(self):
        """
        Force update to image
        """
        MODULE.push_image(self.__view)

    def set_local_rendering(self, local=True):
        self.__app.set(self.__mode_key, "local" if local else "remote")

    def set_remote_rendering(self, remote=True):
        self.__app.set(self.__mode_key, "remote" if remote else "local")

    def update(self):
        # need to do both to keep things in sync
        if self.__rendering:
            self.update_image()
        self.update_geometry()

    def reset_camera(self):
        self.__app.update(ref=self.__ref, method="resetCamera")

    @property
    def view(self):
        """
        Get linked vtkRenderWindow instance
        """
        return self.__wrapped_view


class VtkRemoteView(AbstractElement):
    """
    The VtkRemoteView component relies on the server for rendering by sending images to the client by binding your vtkRenderWindow to it. This component gives you control over the image size and quality to reduce latency while interacting.

    >>> remote_view = vtk.vtkRemoteView(
    ...   view=...,               # Instance of the view (required)
    ...                           # - VTK: vtkRenderWindow
    ...                           # - Paraview: viewProxy
    ...   ref=...,                # Identifier for this component
    ...   interactive_quality=60, # [0, 100] 0 for fastest render, 100 for best quality
    ...   interactive_ratio=...,  # [0.1, 1] Image size scale factor while interacting
    ...   interactor_events=(     # Enable vtk.js interactor events for method binding
    ...     "events",
    ...     ['EndAnimation'],
    ...   ),
    ...   EndAnimation=end,       # Bind method to the enabled event
    ... )
    """

    @staticmethod
    def push_image(view):
        """
        Force image `view` to be pushed to the client
        """
        MODULE.push_image(view)

    def __init__(self, view, ref="view", **kwargs):
        super().__init__("vtk-remote-view", **kwargs)
        self.__app = get_app_instance()
        self.__view = view
        self.__ref = ref
        self._attributes["wsClient"] = ':wsClient="wsClient"'
        self._attributes["ref"] = f'ref="{ref}"'
        self._attributes["view_id"] = f'id="{MODULE.id(view)}"'
        self._attr_names += [
            "enable_picking",
            "interactive_quality",
            "interactive_ratio",
            "interactor_events",
        ]
        self._event_names += kwargs.get("interactor_events", [])

    def update(self):
        """
        Force image to be pushed to client
        """
        MODULE.push_image(self.__view)

    def reset_camera(self):
        self.__app.update(ref=self.__ref, method="resetCamera")


class VtkShareDataset(AbstractElement):
    def __init__(self, children=None, **kwargs):
        super().__init__("vtk-share-dataset", children, **kwargs)
        self._attr_names += ["port", "name"]


class VtkLocalView(AbstractElement):
    """
        The VtkLocalView component relies on the server for defining the vtkRenderWindow but then only the geometry is exchanged with the client. The server does not need a GPU as no rendering is happening on the server. The vtkRenderWindow is only used to retrieve the scene data and parameters (coloring by, representations, ...). By relying on the same vtkRenderWindow, you can easily switch from a ``VtkRemoteView`` to a ``VtkLocalView`` or vice-versa. This component gives you controls on how you want to map mouse interaction with the camera. The default setting mimic default VTK interactor style so you will rarely have to override to the ``interactor_settings``.

    >>> local_view = vtk.VtkLocalView(
    ...   view=...,                # Instance of the view (required)
    ...                            # - VTK: vtkRenderWindow
    ...                            # - Paraview: viewProxy
    ...   ref=...,                 # Identifier for this component
    ...   context_name=...,        # Namespace for geometry cache
    ...   interactor_settings=..., # Options for camera controls. See below.
    ...   interactor_events=(      # Enable vtk.js interactor events for method binding
    ...     "events",
    ...     ['EndAnimation'],
    ...    ),
    ...    EndAnimation=end,       # Bind method to the enabled event
    ... )
    """

    def __init__(self, view, ref="view", **kwargs):
        super().__init__("vtk-sync-view", **kwargs)
        self.__scene_id = f"scene_{ref}"
        self.__view = view
        self.__ref = ref
        self._attributes["ref"] = f'ref="{ref}"'
        self._attributes["wsClient"] = ':wsClient="wsClient"'
        self._attributes["view_state"] = f':viewState="{self.__scene_id}"'
        self._attr_names += ["interactor_events", "interactor_settings", "context_name"]
        self._event_names += kwargs.get("interactor_events", [])
        self.update()

    def update(self):
        """
        Force geometry to be pushed
        """
        _app = get_app_instance()
        _app.set(self.__scene_id, MODULE.scene(self.__view))

    def reset_camera(self):
        """
        Move camera to center actors within the frame
        """
        _app = get_app_instance()
        _app.update(ref=self.__ref, method="resetCamera")


class VtkView(AbstractElement):
    def __init__(self, children=None, ref="view", **kwargs):
        super().__init__("vtk-view", children, **kwargs)
        self._ref = ref
        self._attributes["ref"] = f'ref="{ref}"'
        self._attr_names += [
            "background",
            "cube_axes_style",
            "interactor_settings",
            "picking_modes",
            "show_cube_axes",
        ]
        self._event_names += [
            "hover",
            "click",
            "select",
            "resize",
        ]

    def reset_camera(self):
        """
        Move camera to center actors within the frame
        """
        _app = get_app_instance()
        _app.update(ref=self._ref, method="resetCamera")



from .desktop import AppServerThread, ClientWindowProcess
from .cli import get_cli_parser
from .compose import compose_callbacks
from .filesystem import base_directory
from .logging import log_js_error, print_server_info, validate_key_names
from .string import is_dunder
from .version import get_version


__all__ = [
    "AppServerThread",
    "base_directory",
    "ClientWindowProcess",
    "compose_callbacks",
    "get_cli_parser",
    "get_version",
    "is_dunder",
    "log_js_error",
    "print_server_info",
    "validate_key_names",
]



import asyncio
import os
from pywebvue.utils import read_file_as_base64_url
from trame.html import Span, vuetify, Triggers

import pywebvue
import trame as tr
import trame.internal as tri

LOGO_PATH = os.path.abspath(
    os.path.join(os.path.dirname(__file__), "../html/assets/logo.svg")
)


class AbstractLayout:
    def __init__(self, _root_elem, name, favicon=None, on_ready=None):
        self.name = name
        self.favicon = None
        self.triggers = Triggers("_js_trame_triggers")
        if os.path.exists(LOGO_PATH):
            self.favicon = read_file_as_base64_url(LOGO_PATH)
        self.on_ready = on_ready
        self.children = _root_elem.children
        self._current_root = _root_elem

        # Always add triggers
        self.children += [self.triggers]

        if favicon:
            file_path = os.path.join(tri.base_directory(), favicon)
            if os.path.exists(file_path):
                self.favicon = file_path
            else:
                print(f"Invalid path to favicon: {file_path}")

    @property
    def root(self):
        """
        Top level Vue component. Useful for providing / injecting into children components. Setting makes old root child of new root.
        """
        return self._current_root

    @root.setter
    def root(self, new_root):
        if new_root and self._current_root != new_root:
            new_root.children += [self._current_root]
            self._current_root = new_root

    @property
    def html(self):
        """
        Compute corresponding layout String which represent the html part.
        """
        return self.root.html

    @property
    def state(self):
        """
        Return App state as a dictionary or extend it when setting.
        This is a safe way to build the state incrementaly.

        >>> layout.state = { "a": 1, "b": 2 }
        >>> print(layout.state)
        ... {"a": 1, "b": 2}
        >>> layout.state = { "c": 3, "d": 4 }
        >>> print(layout.state)
        ... {"a": 1, "b": 2, "c": 3, "d": 4}

        """
        return tri.get_app_instance().state

    @state.setter
    def state(self, value):
        _app = tri.get_app_instance()
        for (k, v) in value.items():
            _app.set(k, v)

    def flush_content(self):
        """Push new content to client"""
        _app = tri.get_app_instance()
        _app.layout = self.html

    def start(self, port=None, debug=False):
        """
        Start the application server.

        :param port: Which port to run the server on
        :param debug: Whether to enable debugging tools. Defaults to False.
        :type debug: bool
        """
        _app = tri.get_app_instance()

        _app.name = self.name
        _app.layout = self.html
        _app.on_ready = tri.print_server_info()

        if self.favicon:
            _app.favicon = self.favicon
        if self.on_ready:
            _app.on_ready = tri.print_server_info(self.on_ready)

        # Dev validation
        tri.validate_key_names()

        _app.run_server(port=port)

    def start_thread(
        self, port=None, print_server_info=False, on_server_listening=None, **kwargs
    ):
        _app = tri.get_app_instance()
        _app.name = self.name
        _app.layout = self.html

        if self.favicon:
            _app.favicon = self.favicon

        if print_server_info:
            _app.on_ready = tri.print_server_info(
                tri.compose_callbacks(self.on_ready, on_server_listening)
            )
        else:
            _app.on_ready = tri.compose_callbacks(self.on_ready, on_server_listening)

        # Dev validation
        tri.validate_key_names()
        server_thread = tri.AppServerThread(_app, port, **kwargs)
        server_thread.start()
        return server_thread

    def start_desktop_window(self, on_msg=None, **kwargs):
        from multiprocessing import Queue

        _msg_queue = Queue()

        _app = tri.get_app_instance()
        _app.name = self.name
        _app.layout = self.html

        async def process_msg():
            keep_processing = True
            while keep_processing:
                await asyncio.sleep(0.5)
                if not _msg_queue.empty():
                    msg = _msg_queue.get_nowait()
                    if on_msg:
                        on_msg(msg)
                    if msg == "closing":
                        keep_processing = False
                        _app.stop_server()

        asyncio.get_event_loop().create_task(process_msg())

        if self.favicon:
            _app.favicon = self.favicon

        def start_client(**_):
            client_process = tri.ClientWindowProcess(
                title=_app.name, port=_app.server_port, msg_queue=_msg_queue, **kwargs
            )
            client_process.start()

        _app.on_ready = tri.compose_callbacks(self.on_ready, start_client)

        # Dev validation
        tri.validate_key_names()

        _app.run_server(port=0)


class FullScreenPage(AbstractLayout):
    """
    A layout that takes the whole screen.

    :param name: Text for this page's browser tab (required)
    :type name: str
    :param favicon: Filename of image for this page's browser tab
    :type favicon: str
    :param on_ready: Function to run on startup
    :type on_ready: function

    >>> FullScreenPage("Simple Page").start()
    """

    def __init__(self, name, favicon=None, on_ready=None):
        super().__init__(vuetify.VApp(id="app"), name, favicon, on_ready)


class SinglePage(FullScreenPage):
    """
    A layout that takes the whole screen, adding a |layout_vuetify_link| for a `toolbar`, a VMain as `content` and a VFooter as a `footer`.

    .. |layout_vuetify_link| raw:: html

        <a href="https://vuetifyjs.com/api/v-app-bar" target="_blank">vuetify app bar</a>

    :param name: Text for this page's browser tab (required)
    :type name: str

    >>> layout = SinglePage("Page with header / app bar")

    The toolbar starts with 2 children, a `logo` and a `title` which are accessible at
    the root of the layout object.

    >>> layout.toolbar.children += ["More stuff to the toolbar"]
    >>> layout.logo.children = [VIcon("mdi-menu")]
    >>> layout.title.set_text("My Super App")

    Then we have `content` and `footer`. Content is by default empty but the footer
    has the default trame information regarding its versions and feature feedback
    on when the server is busy with a spining progress.

    You can quickly hide the footer by calling the following.

    >>> layout.footer.hide()
    """

    def __init__(self, name, favicon=None, on_ready=None):
        super().__init__(name, favicon, on_ready)
        self.toolbar = vuetify.VAppBar(app=True)
        if os.path.exists(LOGO_PATH):
            self.logo = Span(
                f'<img height="32px" width="32px" src="{read_file_as_base64_url(LOGO_PATH)}" />',
                classes="mr-2",
                style="display: flex; align-content: center;",
            )
        else:
            self.logo = vuetify.VIcon("mdi-menu", classes="mr-4")

        args = tri.get_cli_parser().parse_known_args()[0]
        dev = args.dev if hasattr(args, "dev") else False

        self.title = Span("trame app", classes="title")
        self.content = vuetify.VMain()
        self.toolbar.children += [self.logo, self.title]
        self.footer = vuetify.VFooter(
            app=True,
            classes="my-0 py-0",
            children=[
                vuetify.VProgressCircular(
                    indeterminate=("busy",),
                    background_opacity=1,
                    bg_color="#01549b",
                    color="#04a94d",
                    size=16,
                    width=3,
                    classes="ml-n3 mr-1",
                ),
                f'<a href="https://kitware.github.io/trame/" class="grey--text lighten-1--text text-caption text-decoration-none" target="_blank">Powered by trame {tr.__version__}/{pywebvue.__version__}</a>',
                vuetify.VSpacer(),
                vuetify.VBtn(
                    vuetify.VIcon("mdi-autorenew", x_small=True),
                    v_if=("__dev_reload", dev),
                    x_small=True,
                    icon=True,
                    click="trigger('server_reload')",
                    classes="mx-2",
                ),
                '<a href="https://www.kitware.com/" class="grey--text lighten-1--text text-caption text-decoration-none" target="_blank">© 2021 Kitware Inc.</a>',
            ],
        )
        self.children += [self.toolbar, self.content, self.footer]


class SinglePageWithDrawer(SinglePage):
    """
    A layout that takes the whole screen, adding a |layout_vuetify_link| for a toolbar, a content, a drawer, and a footer.

    :param name: Text for this page's browser tab (required)
    :type name: str
    :param show_drawer: Whether the drawer is open. Default True
    :type show_drawer: bool
    :param width: How many pixels wide the drawer should be
    :type width: Number
    :param show_drawer_name: The name referencing the drawer's state. Default "drawerOpen".
    :type show_drawer_name: str

    >>> SinglePageWithDrawer("Page with drawer").start()
    """

    def __init__(
        self,
        name,
        favicon=None,
        on_ready=None,
        show_drawer=True,
        width=200,
        show_drawer_name="drawerOpen",
    ):
        super().__init__(name, favicon, on_ready)
        self.drawer = vuetify.VNavigationDrawer(
            app=True,
            clipped=True,
            stateless=True,
            v_model=(show_drawer_name, show_drawer),
            width=width,
        )
        self.toolbar.clipped_left = True
        self.children += [self.drawer]
        self.logo.click = f"{show_drawer_name} = !{show_drawer_name}"


def update_layout(layout):
    """
    Flush layout to the client

    :param layout: UI content for your application
    :type layout: str | trame.layouts.*

    >>> layout.title.set_text("Workload finished!")
    >>> update_layout(layout)

    """
    _app = tri.get_app_instance()
    _app.layout = layout if isinstance(layout, str) else layout.html



from trame.internal import (
  change, Controller, flush_state, get_cli_parser, get_state, get_version,
  is_dirty, is_dirty_all, port, start, State, stop, trigger, update_state
)
from trame.layouts import update_layout

__version__ = get_version()

state = State()
"""This object provides pythonic access to the state

For instance, these getters are the same:

>>> field, = get_state("field")
>>> field = state.field

As are these setters:

>>> update_state("field", value)
>>> state.field = value

``get_state()`` should be used instead if more than one argument is to be
passed, and ``update_state()`` should be used instead to specify additional
arguments (e.g. ``force=True``).

The state may also be accessed and updated similar to dictionaries:

>>> value = state["field"]
>>> state["field"] = value
>>> state.update({"field": value})

This object may be imported via

>>> from trame import state
"""

controller = Controller()
"""The controller is a container for function proxies

The function proxies may be used as callbacks even though the function has
not yet been defined. The function may also be re-defined. For example:

>>> from trame import controller as ctrl
>>> layout = SinglePage("Controller test")
>>> with layout.toolbar:
...     vuetify.VSpacer()
...     vuetify.VBtn("Click Me", click=ctrl.on_click)  # not yet defined

>>> ctrl.on_click = lambda: print("Hello, Trame!")  # on_click is now defined

This can be very useful for large projects where the functions may be defined
in separate files after the UI has been constructed, or for re-defining
callbacks when conditions in the application change.
"""

__all__ = [
    # Order these how we want them to show up in the docs

    # Server-related
    "start",
    "stop",
    "port",

    # State-related
    "state",
    "update_state",
    "get_state",
    "flush_state",
    "is_dirty",
    "is_dirty_all",
    "change",

    # Trigger-related
    "trigger",
    "controller",

    # Layout-related
    "update_layout",

    # CLI-related
    "get_cli_parser",

    # These are not exposed in the docs
    "__version__",
]
